#!/usr/local/bin/perl
use strict;
#------------------------------------------------------------------------------
# Split from Satsuki::DB_pg.pm for AUTOLOAD.
#------------------------------------------------------------------------------
use Satsuki::DB_pg ();
package Satsuki::DB_pg;
our $DEBUG;
###############################################################################
# ơ֥
###############################################################################
#------------------------------------------------------------------------------
# create table
#------------------------------------------------------------------------------
sub create_table {
	my ($self, $table, $columns) = @_;
	my $dbh  = $self->{dbh};
	my $ROBJ = $self->{ROBJ};
	$table =~ s/\W//g;
	if ($table eq '') {  $ROBJ->error('Called create_table() with null table name.'); return 9; }

	# ơ֥빽¤β
	my $cols;
	my @index_columns;
	foreach(@$columns) {
		my $col = $_->{name};
		$col =~ s/\W//g;
		if    ($_->{type} eq 'int')  { $cols .= ", $col INT";     }
		elsif ($_->{type} eq 'flag') { $cols .= ", $col BOOLEAN"; }
		else                         { $cols .= ", $col TEXT";    }
		if ($_->{not_null}) { $cols .= ' NOT NULL'; }		# NOT NULL
		if ($_->{unique})   { $cols .= ' UNIQUE';   }		# ˡ
		elsif ($_->{index}) { push(@index_columns, $col); }	# ˡϼưindex
		if ($_->{ref_pkey}) {	#  table_name.col_name ʸ 
			my $ref_pkey = $_->{ref_pkey};
			$ref_pkey =~ s/[^\w\.]//g;
			my ($ref_table, $ref_col) = split(/\./, $ref_pkey);
			$cols .= " REFERENCES $ref_table($ref_col)";
		}
	}

	# ơ֥
	my $sql = "CREATE TABLE $table(pkey SERIAL PRIMARY KEY$cols)";
	my $sth = $dbh->prepare($sql);
	$DEBUG && $self->debug($sql);	# debug
	$sth && $sth->execute();
	if (!$sth || $dbh->err) { $ROBJ->error($sql); $ROBJ->error($dbh->errstr); return 1; }

	# INDEX 
	#	CREATE INDEX table_colname_idx ON table (colname);
	foreach(@index_columns) {
		my $sql = "CREATE INDEX ${table}_${_}_idx ON $table($_)";
		$dbh->do($sql);
		$DEBUG && $self->debug($sql);	# debug
		if ($dbh->err) { $ROBJ->error($sql); $ROBJ->error($dbh->errstr); return 2; }
	}

	# ơ֥ɲ
	$self->insert_tableinfo($table);

	return 0;
}

#------------------------------------------------------------------------------
# ơ֥κ
#------------------------------------------------------------------------------
#  0 ֤
sub drop_table {
	my ($self, $table) = @_;
	my $dbh  = $self->{dbh};
	my $ROBJ = $self->{ROBJ};
	$table =~ s/\W//g;

	my $sql = "DROP TABLE $table";
	$dbh->do($sql);
	if ($dbh->err) { $ROBJ->error($sql); $ROBJ->error($dbh->errstr); return 1; }

	# ơ֥
	$self->delete_tableinfo($table);
	return 0;
}

###############################################################################
# ǡ
###############################################################################
#------------------------------------------------------------------------------
# ǡ
#------------------------------------------------------------------------------
sub insert {
	my ($self, $table, $h, $save_pkey) = @_;
	my $dbh  = $self->{dbh};
	my $ROBJ = $self->{ROBJ};
	$table =~ s/\W//g;

	my ($cols, $vals);
	my @ary;
	while(my ($k,$v) = each(%$h)) {
		$cols .= "$k,";
		$vals .= '?,';
		if ($v eq '') { undef $v; }
		push(@ary, $v);
	}
	chop($cols); chop($vals);

	# ơ֥򹹿
	$self->update_tableinfo($table);

	# pkey¸ insert
	my $pkey;
	if ($save_pkey) {
		$pkey = int($h->{pkey});
		if ($pkey<1) { return 0; }
		#ʤ Pg Ǥưʤ
		$DEBUG && $self->debug("BEGIN");
		# $dbh->begin_work;
		$DEBUG && $self->debug("LOCK $table SHARE MODE");
		$dbh->do("LOCK $table SHARE MODE");
	} else { delete $h->{pkey}; }

	# SQL ȯ
	my $sql = "INSERT INTO $table($cols) VALUES($vals)";
	my $sth = $dbh->prepare($sql);
	$DEBUG && $self->debug($sql);	# debug
	$sth && $sth->execute(@ary);
	if ($save_pkey && (!$sth || $dbh->err || $sth->rows != 1)) {	# rollback
		$DEBUG && $self->debug("ROLLBACK");
		$ROBJ->error($sql); $ROBJ->error($dbh->errstr);
		$dbh->rollback;
		return 0;
	}
	if (!$sth || $dbh->err) {
		$ROBJ->error($sql); $ROBJ->error($dbh->errstr); return 0;
	}
	if ($sth->rows != 1) { return 0; }

	# pkey¸insertξ硢serialͤꤹ
	if ($save_pkey) {
		$sql = "SELECT setval(pg_catalog.pg_get_serial_sequence('$table', 'pkey'), (SELECT max(pkey) FROM $table))";
		$DEBUG && $self->debug($sql);
		$dbh->do($sql);
		$DEBUG && $self->debug("COMMIT");
		$dbh->commit;
		return $pkey;
	}

	#  pkey ֤
	$sth = $dbh->prepare("SELECT currval('${table}_pkey_seq')");
	$sth && $sth->execute();
	if (!$sth || $dbh->err) { $ROBJ->error($sql); $ROBJ->error($dbh->errstr); return 0; }
	return $sth->fetchrow_array;	# pkey
}

#------------------------------------------------------------------------------
# ǡι
#------------------------------------------------------------------------------
sub update_match {
	my ($self, $table, $h, $match_col, $value, $match_col2, $value2, $match_col3, $value3) = @_;
	my $dbh  = $self->{dbh};
	my $ROBJ = $self->{ROBJ};
	$table      =~ s/\W//g;
	$match_col  =~ s/\W//g;
	$match_col2 =~ s/\W//g;
	$match_col3 =~ s/\W//g;

	# ǡSQL
	my $cols;
	my @ary;
	while(my ($k,$v) = each(%$h)) {
		if ($k eq 'pkey') { next; }
		$cols .= "$k=?,";
		if ($v eq '') { undef $v; }
		push(@ary, $v);
	}
	chop($cols);

	# SQL
	my $where;
	if (ref($value)) {
		foreach(@$value) {
			push(@ary, $_);
			$where .= '?,';
		}
		chop($where);
		if ($where eq '') { return 0; }		# ꤫  ʤ
		$where="$match_col in ($where)";
	} elsif ($match_col ne '') {
		push(@ary, $value);
		$where="$match_col=?";
	}
	# ɲäξ
	if ($match_col2 && $where ne '') {
		push(@ary, $value2);
		$where .= " AND $match_col2=?";
	}
	if ($match_col3 && $where ne '') {
		push(@ary, $value3);
		$where .= " AND $match_col3=?";
	}

	# ơ֥򹹿
	$self->update_tableinfo($table);

	# SQL ȯ
	my $sql = "UPDATE $table SET $cols";
	if ($where ne '') { $sql .= " WHERE $where"; }
	my $sth = $dbh->prepare($sql);
	$DEBUG && $self->debug($sql);	# debug
	$sth && $sth->execute( @ary );
	if (!$sth || $dbh->err) { $ROBJ->error($sql); $ROBJ->error($dbh->errstr); return 0; }

	return $sth->rows;	# Կ
}

#------------------------------------------------------------------------------
# ǡκ
#------------------------------------------------------------------------------
sub delete_match {
	my ($self, $table, $match_col, $value, $match_col2, $value2, $match_col3, $value3) = @_;
	my $dbh  = $self->{dbh};
	my $ROBJ = $self->{ROBJ};
	$table      =~ s/\W//g;
	$match_col  =~ s/\W//g;
	$match_col2 =~ s/\W//g;
	$match_col3 =~ s/\W//g;

	# ι
	my @ary;
	my $where;
	if (ref($value)) {
		foreach(@$value) {
			push(@ary, $_);
			$where .= '?,';
		}
		chop($where);
		if ($where eq '') { return 0; }		# ꤫  ʤ
		$where="$match_col in ($where)";
	} elsif ($match_col ne '') {
		push(@ary, $value);
		$where="$match_col=?";
	}
	# ɲäξ
	if ($match_col2) {
		push(@ary, $value2);
		$where .= " AND $match_col2=?";
	}
	if ($match_col3) {
		push(@ary, $value3);
		$where .= " AND $match_col3=?";
	}

	# ơ֥򹹿
	$self->update_tableinfo($table);

	# SQL ȯ
	my $sql = "DELETE FROM $table";
	if ($where ne '') { $sql .= " WHERE $where"; }
	my $sth = $dbh->prepare($sql);
	$DEBUG && $self->debug($sql);	# debug
	$sth && $sth->execute( @ary );
	if (!$sth || $dbh->err) { $ROBJ->error($sql); $ROBJ->error($dbh->errstr); return 0; }

	return $sth->rows;
}

###############################################################################
# ơν
###############################################################################
#------------------------------------------------------------------------------
# ơ֥ξ򽸷
#------------------------------------------------------------------------------
sub select_by_group {
	my ($self, $table, $h) = @_;
	my $dbh  = $self->{dbh};
	my $ROBJ = $self->{ROBJ};
	$table     =~ s/\W//g;

	my $group_col = $h->{group_col};
	my $sum_col   = $h->{sum_col};
	my $max_col   = $h->{max_col};
	my $min_col   = $h->{min_col};
	my $match_col = $h->{match_col};
	my $match_val = $h->{match_val};
	my $sort      = $h->{sort};
	$match_col =~ s/\W//g;
	$group_col =~ s/\W//g;
	$sum_col   =~ s/\W//g;
	$max_col   =~ s/\W//g;
	$min_col   =~ s/\W//g;
	$sort      =~ s/\W//g;


	#---------------------------------------------
	# Last-Modified ι
	#---------------------------------------------
	$self->update_lastmodifed_by_table( $table );

	#---------------------------------------------
	# SQL
	#---------------------------------------------
	if ($sum_col ne '') {
		$sum_col = ", sum($sum_col) AS _sum";
	}
	if ($max_col ne '') {
		$max_col = ", max($max_col) AS _max";
	}
	if ($min_col ne '') {
		$min_col = ", max($min_col) AS _max";
	}
	if ($match_col ne '') {
		$match_col = " WHERE $match_col=?";
	}
	my $group_by;
	if ($group_col ne '') {
		$group_by  = " GROUP BY $group_col";
		$group_col = ", $group_col";
	}
	if ($sort ne '') {
		$sort = " ORDER BY $sort";
		if ($h->{sort_reverse}) { $sort .= ' DESC'; }
	}

	#---------------------------------------------
	# SQLȯ
	#---------------------------------------------
	my $sql = "SELECT count(pkey) AS _count$sum_col$max_col$min_col$group_col FROM $table$match_col$group_by$sort";
	my $sth = $dbh->prepare($sql);
	$DEBUG && $self->debug($sql);	# debug
	if ($match_col ne '') {
		$sth && $sth->execute($match_val);
	} else {
		$sth && $sth->execute();
	}
	if (!$sth || $dbh->err) { $ROBJ->error($sql); $ROBJ->error($dbh->errstr); return []; }

	return $sth->fetchall_arrayref({});
}

###############################################################################
# ơ֥빹
###############################################################################
#------------------------------------------------------------------------------
# ơ֥빹˥ơ֥ɲ
#------------------------------------------------------------------------------
sub insert_tableinfo {
	my ($self, $table) = @_;
	$self->{no_tableinfo} && return ;
	my $dbh  = $self->{dbh};
	my $ROBJ = $self->{ROBJ};
	my $tableinfo_table = $self->{tableinfo_table};

	# ɲ
	my $sql = "INSERT INTO $tableinfo_table(table_name, tm) VALUES(?,?)";
	my $r   = $self->do_sql(0, $sql, $table, $ROBJ->{TM});
	if ($r>0) {
		# ơ֥Τιѹ
		$self->update_tableinfo( $tableinfo_table );
		return $r;
	}

	# Ԥơ֥κߤ
	$self->create_tableinfo_table( $tableinfo_table );
	return $self->do_sql($sql, $table, $ROBJ->{TM});
}

#------------------------------------------------------------------------------
# ơ֥빹ν񤭴
#------------------------------------------------------------------------------
sub update_tableinfo {
	my ($self, $table) = @_;
	$self->{no_tableinfo} && return ;
	my $tableinfo_table = $self->{tableinfo_table};
	my $cache_id = $self->{db_id} . $table;
	$self->{lastmodified_cache}->{$cache_id} = $self->{ROBJ}->{TM};
	return $self->do_sql("UPDATE $tableinfo_table SET tm=? WHERE table_name=?", $self->{ROBJ}->{TM}, $table);
}
#------------------------------------------------------------------------------
# ơ֥빹κ
#------------------------------------------------------------------------------
sub delete_tableinfo {
	my ($self, $table) = @_;
	$self->{no_tableinfo} && return ;
	# å
	delete $self->{lastmodified_cache}->{ $self->{db_id} . $table };
	# ơ֥Τιѹ
	my $tableinfo_table = $self->{tableinfo_table};
	$self->update_tableinfo( $tableinfo_table );
	# 
	return $self->do_sql("DELETE FROM $tableinfo_table WHERE table_name=?", $table);
}
#------------------------------------------------------------------------------
# ơ֥빹ơ֥κ
#------------------------------------------------------------------------------
sub create_tableinfo_table {
	my ($self, $tableinfo_table) = @_;
	$self->{no_tableinfo} && return ;
	$self->do_sql("CREATE TABLE $tableinfo_table(pkey SERIAL PRIMARY KEY, table_name TEXT UNIQUE NOT NULL, tm INT)");
	$self->do_sql("CREATE INDEX ${tableinfo_table}_table_name_idx ON $tableinfo_table(table_name)");
	$self->do_sql("INSERT INTO $tableinfo_table(table_name, tm) VALUES('$tableinfo_table', ?)", $self->{ROBJ}->{TM});
}

#------------------------------------------------------------------------------
# SQLμ¹
#------------------------------------------------------------------------------
sub do_sql {
	my $self  = shift;
	my $sql   = shift;
	my $err_f = 1;						# 顼ɽե饰
	if ($sql =~ /^\d$/) { $err_f=$sql; $sql=shift; }	# 顼ɽʤ
	my $dbh   = $self->{dbh};
	$DEBUG && $self->debug($sql);	# debug
	my $sth = $dbh->prepare($sql);
	$sth && $sth->execute(@_);
	if ($err_f && (!$sth || $dbh->err)) { $self->{ROBJ}->error($sql); $self->{ROBJ}->error($dbh->errstr); }
	return $sth && $sth->rows;
}

1;	# ޤʤ
